/*
 * linux/include/asm-arm/arch-ebsa110/timex.h
 *
 * RiscPC architecture timex specifications
 *
 * Copyright (C) 1997, 1998 Russell King
 */

/*
 * On the EBSA, the clock ticks at weird rates.
 * This is therefore not used to calculate the
 * divisor.
 */
//#define CLOCK_TICK_RATE		2000000

