/* SF16FMI FMRadio include file.
 * (c) 1998 Petr Vandrovec
 *
 * Not in include/linux/ because there's no need for anyone
 * to know about these details, I reckon.
 */

#ifndef __RSF16FMI_H
#define __RSF16FMI_H

#include <linux/radio.h>

int radiosf16fmi_init(void);

#endif /* __RSF16FMI_H */
